from abc import ABC, abstractmethod
from pyioc3 import StaticContainerBuilder, ScopeEnum
import unittest


class Interface(ABC):
    @abstractmethod
    def method(self) -> None:
        raise NotImplementedError()


class DependentInterface(ABC):
    @abstractmethod
    def x(self) -> None:
        raise NotImplementedError()


class CacheInterface(ABC):
    @abstractmethod
    def inc(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get_count(self) -> int:
        raise NotImplementedError()


class Cache(CacheInterface):
    def __init__(self):
        self._count = 0

    def inc(self) -> None:
        self._count += 1

    def get_count(self) -> None:
        return self._count


class Cache2(CacheInterface):
    def __init__(self):
        self._count = 0

    def inc(self) -> None:
        self._count += 1

    def get_count(self) -> None:
        return self._count


class Parent(Interface):
    def __init__(self, dep: DependentInterface):
        self._dep = dep

    def method(self):
        self._dep.x()


class Child(DependentInterface):
    def __init__(self, cache: CacheInterface):
        self.cache = cache

    def x(self):
        self.cache.inc()


class SystemTest(unittest.TestCase):
    def test_simple(self):
        builder = StaticContainerBuilder()
        builder.bind(Interface, Parent)
        builder.bind(DependentInterface, Child)
        builder.bind(CacheInterface, Cache, ScopeEnum.SINGLETON)
        ioc = builder.build()
        parent = ioc.get(Interface)
        cache = ioc.get(CacheInterface)
        parent.method()
        self.assertEqual(cache.get_count(), 1)

    def test_simple_with_str_scope(self):
        builder = StaticContainerBuilder()
        builder.bind(Interface, Parent)
        builder.bind(DependentInterface, Child)
        builder.bind(CacheInterface, Cache, "SINGLETON")
        ioc = builder.build()
        parent = ioc.get(Interface)
        cache = ioc.get(CacheInterface)
        parent.method()
        self.assertEqual(cache.get_count(), 1)



import unittest

from pyioc3 import ScopeEnum
from pyioc3.errors import ScopeError


class SystemTest(unittest.TestCase):
    def test_transient_from_string(self):
        assert ScopeEnum.TRANSIENT == ScopeEnum.from_string("transient")
        assert ScopeEnum.TRANSIENT == ScopeEnum.from_string("t")
        assert ScopeEnum.TRANSIENT == ScopeEnum.from_string("TRANSIENT")
        assert ScopeEnum.TRANSIENT == ScopeEnum.from_string("T")

    def test_requested_from_string(self):
        assert ScopeEnum.REQUESTED == ScopeEnum.from_string("requested")
        assert ScopeEnum.REQUESTED == ScopeEnum.from_string("r")
        assert ScopeEnum.REQUESTED == ScopeEnum.from_string("REQUESTED")
        assert ScopeEnum.REQUESTED == ScopeEnum.from_string("R")

    def test_singleton_from_string(self):
        assert ScopeEnum.SINGLETON == ScopeEnum.from_string("singleton")
        assert ScopeEnum.SINGLETON == ScopeEnum.from_string("s")
        assert ScopeEnum.SINGLETON == ScopeEnum.from_string("SINGLETON")
        assert ScopeEnum.SINGLETON == ScopeEnum.from_string("S")

    def test_invalid_scope_raises(self):
        with self.assertRaises(ScopeError):
            ScopeEnum.from_string("foo")



from pyioc3.errors import CircularDependencyError, MemberNotBoundError
from pyioc3.interface import Container
from pyioc3.static_container_builder import StaticContainerBuilder
from unittest.mock import patch
import unittest

from .fixtures import (
    DuckA,
    DuckInterface,
    HalfCircle1,
    HalfCircle2,
    QuackBehavior,
    Sqeak,
)


class StaticContainerTest(unittest.TestCase):
    def setUp(self):
        self.builder = StaticContainerBuilder()

    def test_raises_if_deep_cycle(self):
        self.builder.bind(HalfCircle1, HalfCircle1)
        self.builder.bind(HalfCircle2, HalfCircle2)
        with self.assertRaises(CircularDependencyError):
            self.builder.build()

    def test_raises_if_any_cycle(self):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.bind(HalfCircle1, HalfCircle1)
        self.builder.bind(HalfCircle2, HalfCircle2)
        with self.assertRaises(CircularDependencyError):
            self.builder.build()

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_builds_container(self, container_mock):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        container_mock.assert_called()

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_builds_container_with_initialized_dep_graph(self, container_mock):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        members, *_ = container_mock.call_args[0]
        (dep,) = list(members[DuckInterface])
        self.assertEqual(QuackBehavior, dep.annotation)

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_builds_container_with_members(self, container_mock):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        members, *_ = container_mock.call_args[0]
        self.assertIn(DuckInterface, members)
        self.assertIn(QuackBehavior, members)

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_builds_with_static_container(self, container_mock):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        members, *_ = container_mock.call_args[0]
        self.assertIn(container_mock, members)

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_builds_with_static_container_as_container_interface(self, container_mock):
        self.builder.bind(DuckInterface, DuckA)
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        members, *_ = container_mock.call_args[0]
        self.assertIn(Container, members)

    @patch("pyioc3.static_container_builder.StaticContainer")
    def test_binding_scope_by_string_is_ok(self, container_mock):
        self.builder.bind(DuckInterface, DuckA, "singleton")
        self.builder.bind(QuackBehavior, Sqeak)
        self.builder.build()
        members, *_ = container_mock.call_args[0]
        self.assertIn(Container, members)

    def test_unbound_dependent_raises_key_error(self):
        self.builder.bind(DuckInterface, DuckA, "singleton")
        with self.assertRaises(KeyError):
            self.builder.build()

    def test_unbound_dependent_raises_member_not_bound_error(self):
        self.builder.bind(DuckInterface, DuckA, "singleton")
        with self.assertRaises(MemberNotBoundError):
            self.builder.build()



import unittest
from unittest.mock import MagicMock

from pyioc3.errors import MemberNotBoundError
from pyioc3.static_container import StaticContainer
from pyioc3.scope_enum import ScopeEnum
from pyioc3.bound_member import BoundMember


class StaticContainerTest(unittest.TestCase):
    def setUp(self):
        foo1 = BoundMember(
            annotation="foo1",
            implementation=object,
            scope=ScopeEnum.TRANSIENT,
            parameters=[],
        )

        foo2 = BoundMember(
            annotation="foo2",
            implementation=MagicMock(),
            scope=ScopeEnum.TRANSIENT,
            parameters=["foo1"],
        )

        foo3 = BoundMember(
            annotation="foo3",
            implementation=MagicMock(),
            scope=ScopeEnum.TRANSIENT,
            parameters=["foo1"],
        )

        foo2.bind_dependant(foo1)
        foo3.bind_dependant(foo1)

        self.members = {f.annotation: f for f in [foo1, foo2, foo3]}
        self.container = StaticContainer(self.members)

    def test_retrieves_instance(self):
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        bar = self.container.get("foo1")
        self.assertEqual("bar", bar)

    def test_retrieves_unique_instance_when_transient(self):
        obj1 = self.container.get("foo1")
        obj2 = self.container.get("foo1")
        self.assertTrue(obj1 is not obj2)

    def test_retrieves_unique_instance_when_requested(self):
        self.members["foo1"].scope = ScopeEnum.REQUESTED
        obj1 = self.container.get("foo1")
        obj2 = self.container.get("foo1")
        self.assertTrue(obj1 is not obj2)

    def test_retrieves_same_instance_when_singleton(self):
        self.members["foo1"].scope = ScopeEnum.SINGLETON
        obj1 = self.container.get("foo1")
        obj2 = self.container.get("foo1")
        self.assertTrue(obj1 is obj2)

    def test_injects_deps(self):
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.container.get("foo2")
        self.members["foo2"].implementation.assert_called_with("bar")

    def test_injected_deps_are_unique_when_transient(self):
        self.container.get("foo2")
        self.container.get("foo3")
        self.assertNotEqual(
            self.members["foo2"].implementation.call_args,
            self.members["foo3"].implementation.call_args,
        )

    def test_injected_deps_are_the_same_when_singleton(self):
        self.members["foo1"].scope = ScopeEnum.SINGLETON
        self.container.get("foo2")
        self.container.get("foo3")
        self.assertEqual(
            self.members["foo2"].implementation.call_args,
            self.members["foo3"].implementation.call_args,
        )

    def test_injected_deps_are_unqueue_when_requested(self):
        self.members["foo1"].scope = ScopeEnum.REQUESTED
        self.container.get("foo2")
        self.container.get("foo3")
        self.assertNotEqual(
            self.members["foo2"].implementation.call_args,
            self.members["foo3"].implementation.call_args,
        )

    def test_injected_deps_are_the_same_when_requested_in_same_tree(self):
        self.members["foo1"].scope = ScopeEnum.REQUESTED
        self.members["foo3"].parameters = ["foo1", "foo2"]
        self.members["foo3"].bind_dependant(self.members["foo2"])
        self.container.get("foo3")
        foo2_args = self.members["foo2"].implementation.call_args
        foo3_args = self.members["foo3"].implementation.call_args
        self.assertEqual(foo2_args[0][0], foo3_args[0][0])

    def test_calls_on_activate(self):
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.members["foo1"].on_activate = MagicMock(return_value="bar")
        self.container.get("foo1")
        self.members["foo1"].on_activate.assert_called_with("bar")

    def test_calls_on_activate_for_dep(self):
        self.members["foo1"].scope = ScopeEnum.SINGLETON
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.members["foo1"].on_activate = MagicMock(return_value="bar")
        self.container.get("foo2")
        self.members["foo1"].on_activate.assert_called_with("bar")

    def test_calls_on_activate_only_once_when_singleton(self):
        self.members["foo1"].scope = ScopeEnum.SINGLETON
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.members["foo1"].on_activate = MagicMock(return_value="bar")
        self.container.get("foo2")
        self.container.get("foo3")
        self.members["foo1"].on_activate.assert_called_once_with("bar")

    def test_calls_on_activate_only_once_when_requested_in_same_tree(self):
        self.members["foo1"].scope = ScopeEnum.REQUESTED
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.members["foo1"].on_activate = MagicMock(return_value="bar")
        self.members["foo3"].bind_dependant(self.members["foo2"])
        self.container.get("foo3")
        self.members["foo1"].on_activate.assert_called_once_with("bar")

    def test_calls_on_activate_each_time_it_is_needed_when_transient(self):
        self.members["foo1"].scope = ScopeEnum.TRANSIENT
        self.members["foo1"].implementation = MagicMock(return_value="bar")
        self.members["foo1"].on_activate = MagicMock(return_value="bar")
        self.members["foo3"].bind_dependant(self.members["foo2"])
        self.container.get("foo3")
        self.assertEqual(2, self.members["foo1"].on_activate.call_count)

    def test_unbound_member_raises_key_error(self):
        with self.assertRaises(KeyError):
            self.container.get("NoExist")

    def test_unbound_member_raises_member_not_bound_error(self):
        with self.assertRaises(MemberNotBoundError):
            self.container.get("NoExist")



import unittest
from unittest.mock import MagicMock, patch

from pyioc3.scope_container import ScopeContainer, PersistentScope
from pyioc3.scope_enum import ScopeEnum
from pyioc3.bound_member import BoundMember


class ContainerScopeTests(unittest.TestCase):
    @patch("pyioc3.scope_container.PersistentScope")
    @patch("pyioc3.scope_container.TransientScope")
    def setUp(self, transient, persistent):
        self.transient = transient()
        self.persistent = persistent()
        self.singleton = persistent()
        self.container = ScopeContainer(self.singleton)

        self.scope_cases = [
            (ScopeEnum.TRANSIENT, self.transient),
            (ScopeEnum.REQUESTED, self.persistent),
            (ScopeEnum.SINGLETON, self.singleton),
        ]

    def test_add_uses_correct_scope(self):
        for scope_id, scope in self.scope_cases:
            with self.subTest(msg=scope_id):
                member = MagicMock(spec=BoundMember)
                member.scope = scope_id
                member.annotation = "foo"
                member.implementation = MagicMock(return_value="bar")
                member.on_activate = lambda x: x
                self.container.add(member)
                scope.add.assert_called_with("foo", "bar")

    def test_has_uses_correct_scope(self):
        for scope_id, scope in self.scope_cases:
            with self.subTest(msg=scope_id):
                member = MagicMock(spec=BoundMember)
                member.scope = scope_id
                member.annotation = "foo"
                member.implementation = MagicMock(return_value="bar")
                member.on_activate = lambda x: x
                self.container.has(member)
                scope.__contains__.assert_called_with("foo")

    def test_get_instance_of_uses_correct_scope(self):
        for scope_id, scope in self.scope_cases:
            with self.subTest(msg=scope_id):
                member = MagicMock(spec=BoundMember)
                member.scope = scope_id
                member.annotation = "foo"
                member.implementation = MagicMock(return_value="bar")
                member.on_activate = lambda x: x
                self.container.get_instance_of(member)
                scope.use.assert_called_with("foo")


class ScopeContainerTests(unittest.TestCase):
    def setUp(self):
        self.container = ScopeContainer(PersistentScope())

    def test_creates_inst_with_deps(self):
        a1 = BoundMember(
            annotation="a1",
            implementation=lambda: "a1_impl",
            scope=ScopeEnum.SINGLETON,
            parameters=[],
        )
        a1._depends_on = []
        self.container.add(a1)

        a2 = BoundMember(
            annotation="a2",
            implementation=lambda: "a2_impl",
            scope=ScopeEnum.SINGLETON,
            parameters=[],
        )
        a2._depends_on = []
        self.container.add(a2)

        impl = MagicMock()
        member = BoundMember(
            annotation="foo",
            implementation=impl,
            scope=ScopeEnum.SINGLETON,
            parameters=["a1", "a2"],
        )
        member._depends_on = [a1, a2]
        self.container.add(member)

        impl.assert_called_with("a1_impl", "a2_impl")

    def test_on_activate_called(self):
        on_activate = MagicMock(return_value="baz")
        member = BoundMember(
            annotation="foo",
            implementation=lambda: "bar",
            scope=ScopeEnum.SINGLETON,
            parameters=["a1", "a2"],
            on_activate=on_activate,
        )
        member._depends_on = []
        self.container.add(member)
        on_activate.assert_called_with("bar")

    def test_on_activate_called_once_if_singleton(self):
        on_activate = MagicMock(return_value="baz")

        a1 = BoundMember(
            annotation="a1",
            implementation=lambda: "a1_impl",
            scope=ScopeEnum.SINGLETON,
            parameters=[],
            on_activate=on_activate,
        )
        a1._depends_on = []

        a2 = BoundMember(
            annotation="a2",
            implementation=lambda a: "a2_impl",
            scope=ScopeEnum.SINGLETON,
            parameters=["a1"],
        )
        a2._depends_on = [a1]

        member = BoundMember(
            annotation="foo",
            implementation=lambda a, b: "bar",
            scope=ScopeEnum.SINGLETON,
            parameters=["a1", "a2"],
        )
        member._depends_on = [a1, a2]

        self.container.add(a1)
        self.container.add(a2)
        self.container.add(member)

        on_activate.assert_called_once_with("a1_impl")

    def test_on_activate_called_for_each_instance_if_transient(self):
        on_activate = MagicMock(return_value="baz")

        a1 = BoundMember(
            annotation="a1",
            implementation=lambda: "a1_impl",
            scope=ScopeEnum.TRANSIENT,
            parameters=[],
            on_activate=on_activate,
        )
        a1._depends_on = []

        a2 = BoundMember(
            annotation="a2",
            implementation=lambda a: "a2_impl",
            scope=ScopeEnum.SINGLETON,
            parameters=["a1"],
        )
        a2._depends_on = [a1]

        member = BoundMember(
            annotation="foo",
            implementation=lambda a, b: "bar",
            scope=ScopeEnum.SINGLETON,
            parameters=["a1", "a2"],
        )
        member._depends_on = [a1, a2]

        self.container.add(a1)
        self.container.add(a1)
        self.container.add(a2)
        self.container.add(member)

        self.assertEqual(2, on_activate.call_count)



import unittest
from random import randint, choice
from pyioc3.interface import Container, ProviderBinding
from pyioc3.bound_member_factory import BoundMemberFactory
from pyioc3.adapters import (
    ValueAsImplAdapter,
    FunctionAsImplAdapter,
    FactoryAsImplAdapter,
)

from .fixtures import (
    DuckA,
    DuckB,
    DuckC,
    duck_d,
    DuckInterface,
    QuackBehavior,
    FlockOfDucks,
    FlockInterface,
    MasterDuck,
    MetaMasterDuck,
)


class BoundMemberTest(unittest.TestCase):
    def test_expands_parameter_annotations_for_classes(self):
        cases = [
            (ProviderBinding(DuckInterface, DuckA), [QuackBehavior]),
            (ProviderBinding(DuckInterface, DuckB), [QuackBehavior]),
            (ProviderBinding(DuckInterface, DuckC), []),
            (ProviderBinding("foo", duck_d), [QuackBehavior]),
            (
                ProviderBinding(FlockInterface[DuckInterface], FlockOfDucks),
                [DuckA, DuckB],
            ),
            (ProviderBinding("lambda", lambda: "x"), []),
            (ProviderBinding("value", ValueAsImplAdapter(1)), []),
            (ProviderBinding("function", FunctionAsImplAdapter(lambda: "x")), []),
            (
                ProviderBinding("factory", FactoryAsImplAdapter(lambda: "x")),
                [Container],
            ),
            (ProviderBinding(DuckInterface, MasterDuck), [QuackBehavior]),
            (ProviderBinding(DuckInterface, MetaMasterDuck), [QuackBehavior]),
        ]
        for binding, result in cases:
            with self.subTest(binding):
                b = BoundMemberFactory.build(binding)
                self.assertListEqual(b.parameters, result)

    def test_ordering_of_parameter_annotations(self):
        types = {
            "int": int,
            "str": str,
            "bool": bool,
        }
        params = []
        for i in range(255):
            val = f"p{randint(1, 1000)}_{i}"
            anno = choice(list(types.keys()))
            params.append((val, anno))
        args = ",".join([f"{p}: {a}" for p, a in params])
        expected = [types[a] for _, a in params]
        g = {}
        exec(f"def func({args}):...", g)
        b = BoundMemberFactory.build(ProviderBinding("foo", g["func"]))
        self.assertListEqual(expected, b.parameters)



from unittest import TestCase
from pyioc3 import ScopeEnum
from pyioc3.errors import AutoWireError
from pyioc3.autowire import bind, bind_factory, AutoWireContainerBuilder
from pyioc3.interface import ProviderBinding, FactoryBinding


class AutoWireDecoratorTests(TestCase):
    def setUp(self):
        AutoWireContainerBuilder._staged_bindings = []

    def test_binding_decorator_stages_binding_with_defaults(self):
        @bind()
        class MyClass:
            ...

        self.assertListEqual(
            AutoWireContainerBuilder._staged_bindings,
            [
                (
                    "tests.test_autowire",
                    ProviderBinding(MyClass, MyClass, ScopeEnum.TRANSIENT, None),
                ),
            ],
        )

    def test_binding_decorator_sets_scope(self):
        @bind(scope="SINGLETON")
        class MyClass:
            ...

        self.assertListEqual(
            AutoWireContainerBuilder._staged_bindings,
            [
                (
                    "tests.test_autowire",
                    ProviderBinding(MyClass, MyClass, "SINGLETON", None),
                )
            ],
        )

    def test_binding_decorator_sets_annotation(self):
        class MyClassInterface:
            ...

        @bind(MyClassInterface)
        class MyClass:
            ...

        self.assertListEqual(
            AutoWireContainerBuilder._staged_bindings,
            [
                (
                    "tests.test_autowire",
                    ProviderBinding(
                        MyClassInterface, MyClass, ScopeEnum.TRANSIENT, None
                    ),
                )
            ],
        )

    def test_binding_decorator_stages_binding_with_on_activate(self):
        def my_activator(x):
            return x

        @bind(on_activate=my_activator)
        class MyClass:
            ...

        self.assertListEqual(
            AutoWireContainerBuilder._staged_bindings,
            [
                (
                    "tests.test_autowire",
                    ProviderBinding(
                        MyClass, MyClass, ScopeEnum.TRANSIENT, my_activator
                    ),
                )
            ],
        )

    def test_binding_factory_decorator_stages_binding_with_defaults(self):
        @bind_factory("MyFactory")
        def my_factory(ctx):
            def wrapped(foo):
                ...

            return wrapped

        self.assertListEqual(
            AutoWireContainerBuilder._staged_bindings,
            [
                (
                    "tests.test_autowire",
                    FactoryBinding(my_factory, "MyFactory"),
                )
            ],
        )

    def test_binding_factory_requires_annotation(self):
        with self.assertRaises(TypeError):

            @bind_factory()
            def my_factory(ctx):
                def wrapped(foo):
                    ...

                return wrapped


class AutoWireDiscoveryTests(TestCase):
    def setUp(self):
        AutoWireContainerBuilder._staged_bindings = []

    def test_autowire_can_build_simple_module(self):
        output = []

        class QuackInterface:
            def quack(self) -> None:
                raise NotImplementedError()

        @bind(QuackInterface)
        class Squeak(QuackInterface):
            def quack(self) -> None:
                output.append("Squeak")

        @bind()
        class Duck:
            def __init__(self, quack: QuackInterface) -> None:
                self._quack = quack

            def quack(self) -> None:
                self._quack.quack()

        container = AutoWireContainerBuilder("tests.test_autowire").build()
        duck = container.get(Duck)
        self.assertIsInstance(duck, Duck)
        duck.quack()
        self.assertListEqual(output, ["Squeak"])

    def test_autowire_can_build_from_submodules(self):
        from tests.autowire_pkg.greeter import GreeterOpts
        from tests.autowire_pkg.interface import GreeterInterface

        container = (
            AutoWireContainerBuilder("tests.autowire_pkg", "tests.autowire_pkg.script")
            .bind_constant(GreeterOpts, GreeterOpts(default_name="World"))
            .build()
        )
        greeter = container.get(GreeterInterface)
        self.assertEqual(greeter.greet("Unittest"), "Hello, Unittest!")
        self.assertEqual(greeter.greet(), "Hello, World!")
        greeter.set_language("es_MX")
        self.assertEqual(greeter.greet("Ian"), "Hola, Ian!")
        self.assertEqual(greeter.greet(), "Hola, Mundo!")

    def test_autowire_raises_on_duplicates(self):
        class QuackInterface:
            def quack(self) -> None:
                ...

        @bind(QuackInterface)
        class Squeak(QuackInterface):
            def quack(self) -> None:
                ...

        @bind(QuackInterface)
        class Honk(QuackInterface):
            def quack(self) -> None:
                ...

        @bind()
        class Duck:
            def __init__(self, quack: QuackInterface) -> None:
                ...

        with self.assertRaises(AutoWireError):
            AutoWireContainerBuilder("tests.test_autowire").build()



import unittest
from typing import Callable

from pyioc3.builder import BuilderBase
from pyioc3.interface import ProviderBinding, ConstantBinding, FactoryBinding


GreeterFactory = Callable[[str], str]


class Foo:
    ...


class FooBuilder(BuilderBase[Foo]):
    def __init__(self, *args, **kwargs):
        super().__init__(Foo, *args, **kwargs)


def eng_greeter_factory(ctx) -> GreeterFactory:
    def greeter_factory(name: str) -> str:
        return f"Hello, {name}!"

    return greeter_factory


class BuilderTest(unittest.TestCase):
    def test_builder(self):
        foo = FooBuilder().build()
        assert isinstance(foo, Foo)

    def test_builder_with_provider_and_constant_defaults(self):
        class Foo_Test(Foo):
            def __init__(self, c: str):
                self.c = c

        foo = FooBuilder(
            [
                ProviderBinding(Foo, Foo_Test),
                ConstantBinding("bar", str),
            ]
        ).build()
        assert foo.c == "bar"

    def test_builder_with_provider_and_constant_and_factory_defaults(self):
        class Foo_Test(Foo):
            def __init__(self, name: str, greet: GreeterFactory):
                self.greeting = greet(name)

        foo = FooBuilder(
            [
                ProviderBinding(Foo, Foo_Test),
                ConstantBinding("world", str),
                FactoryBinding(eng_greeter_factory, GreeterFactory),
            ]
        ).build()

        assert foo.greeting == "Hello, world!"

    def test_builder_with_provider_override(self):
        foo = FooBuilder().using_provider(Foo).build()
        assert isinstance(foo, Foo)

    def test_builder_with_provider_and_constant(self):
        class Foo_Test(Foo):
            def __init__(self, name: str):
                self.name = name

        foo = (
            FooBuilder()
            .using_provider(Foo, Foo_Test)
            .using_constant(str, "World")
            .build()
        )

        assert foo.name == "World"

    def test_builder_with_provider_and_constant_and_factory_override(self):
        class Foo_Test(Foo):
            def __init__(self, name: str, greet: GreeterFactory):
                self.greeting = greet(name)

        foo = (
            FooBuilder()
            .using_provider(Foo, Foo_Test)
            .using_constant(str, "World")
            .using_factory(GreeterFactory, eng_greeter_factory)
            .build()
        )

        assert foo.greeting == "Hello, World!"

    def test_mixed(self):
        class Bar(Foo):
            def __init__(self, name: str, greet: GreeterFactory):
                self.greeting = greet(name)

        class FooTestBuilder(BuilderBase[Bar]):
            def __init__(self):
                super().__init__(Bar)

        bar = (
            FooTestBuilder()
            .using_constant(str, "World")
            .using_factory(GreeterFactory, eng_greeter_factory)
            .build()
        )

        assert isinstance(bar, Bar)
        assert bar.greeting == "Hello, World!"



import unittest

from pyioc3.scope_container import PersistentScope


class PersistentScopeTest(unittest.TestCase):
    def setUp(self):
        self.scope = PersistentScope()

    def test_only_once_instance_exists(self):
        self.scope.add("a", 1)
        self.scope.add("a", 2)
        self.scope.add("a", 3)
        self.assertEqual(3, self.scope.use("a"))
        self.assertEqual(3, self.scope.use("a"))
        self.assertEqual(3, self.scope.use("a"))

    def test_raises_key_error(self):
        self.scope.add("a", 1)
        with self.assertRaises(KeyError):
            self.scope.use("b")

    def test_in_oper_returns_true(self):
        self.scope.add("a", 1)
        self.assertIn("a", self.scope)

    def test_in_oper_returns_false(self):
        self.scope.add("a", 1)
        self.assertNotIn("b", self.scope)

    def test_item_is_not_removed_once_used(self):
        self.scope.add("a", 1)
        self.scope.use("a")
        self.scope.use("a")
